/*
 *
 *  kernel/cpt/cpt_kernel.c
 *
 *  Copyright (C) 2000-2005  SWsoft
 *  All rights reserved.
 *
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 */

#define __KERNEL_SYSCALLS__ 1

#include <linux/version.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <asm/cpufeature.h>
#include <linux/cpt_image.h>

#include "cpt_kernel.h"
#include "cpt_syscalls.h"

int debug_level = 1;

#ifndef CONFIG_X86_64

extern void local_kernel_thread_helper(void);
__asm__(".section .text\n"
	".align 4\n"
	"local_kernel_thread_helper:\n\t"
	"movl %edx,%eax\n\t"
	"pushl %edx\n\t"
	"call *%ebx\n\t"
	"pushl %eax\n\t"
	"pushl $0\n\t"
	"call complete_and_exit\n"
	".previous");

/*
 * Create a kernel thread
 */
int asm_kernel_thread(int (*fn)(void *), void * arg, unsigned long flags, pid_t pid)
{
	struct pt_regs regs;

	memset(&regs, 0, sizeof(regs));

	regs.ebx = (unsigned long) fn;
	regs.edx = (unsigned long) arg;

	regs.xds = __USER_DS;
	regs.xes = __USER_DS;
	regs.orig_eax = -1;
	regs.eip = (unsigned long) local_kernel_thread_helper;
	regs.xcs = __KERNEL_CS;
	regs.eflags = X86_EFLAGS_IF | X86_EFLAGS_SF | X86_EFLAGS_PF | 0x2;

	/* Ok, create the new process.. */
	return do_fork_pid(flags | CLONE_UNTRACED, 0, &regs, 0, NULL, NULL, pid);
}
#endif

int local_kernel_thread(int (*fn)(void *), void * arg, unsigned long flags, pid_t pid)
{
	pid_t ret;

	if (!try_module_get(THIS_MODULE))
		return -EBUSY;
	ret = asm_kernel_thread(fn, arg, flags, pid);
	if (ret < 0)
		module_put(THIS_MODULE);
	return ret;
}

#ifdef __i386__
int __execve(const char *file, char **argv, char **envp)
{
	long res;
	__asm__ volatile ("int $0x80"
	: "=a" (res)
	: "0" (__NR_execve),"b" ((long)(file)),"c" ((long)(argv)),
		  "d" ((long)(envp)) : "memory");
	return (int)res;
}
#endif

int sc_execve(char *cmd, char **argv, char **env)
{
	int ret;
#ifndef __i386__
	ret = execve(cmd, argv, env);
#else
	ret = __execve(cmd, argv, env);
#endif
	return ret;
}

unsigned int test_cpu_caps()
{
	unsigned int flags = 0;
	if (boot_cpu_has(X86_FEATURE_CMOV))
		flags |= 1 << CPT_CPU_X86_CMOV;
	if (cpu_has_fxsr)
		flags |= 1 << CPT_CPU_X86_FXSR;
	if (cpu_has_xmm)
		flags |= 1 << CPT_CPU_X86_SSE;
#ifndef CONFIG_X86_64
	if (cpu_has_xmm2)
#endif
		flags |= 1 << CPT_CPU_X86_SSE2;
	if (cpu_has_mmx)
		flags |= 1 << CPT_CPU_X86_MMX;
	if (boot_cpu_has(X86_FEATURE_3DNOW))
		flags |= 1 << CPT_CPU_X86_3DNOW;
	if (boot_cpu_has(X86_FEATURE_3DNOWEXT))
		flags |= 1 << CPT_CPU_X86_3DNOW2;
	if (boot_cpu_has(X86_FEATURE_SEP))
		flags |= 1 << CPT_CPU_X86_SEP;
#ifdef CONFIG_X86_64
	flags |= 1 << CPT_CPU_X86_EMT64;
#endif
	return flags;
}

unsigned int test_kernel_config()
{
	unsigned int flags = 0;
#if defined(CONFIG_X86_PAE) || defined(CONFIG_X86_64)
	flags |= 1 << CPT_KERNEL_CONFIG_PAE;
#endif
	return flags;
}
