/* Interface to kernel vars which we had to _add_. */

asmlinkage void ret_from_fork(void) __asm__("ret_from_fork");

#define PRIO_TO_NICE(prio)	((prio) - MAX_RT_PRIO - 20)

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
#define TASK_TRACED TASK_STOPPED
#define unix_peer(sk) ((sk)->sk_pair)
#define page_mapcount(pg) ((pg)->mapcount)
#else
#define unix_peer(sk) (unix_sk(sk)->peer)
#endif

#ifdef CONFIG_X86_64
#define cpu_has_fxsr 1
#endif

static inline void do_gettimespec(struct timespec *ts)
{
	struct timeval tv;
	do_gettimeofday(&tv);
	ts->tv_sec = tv.tv_sec;
	ts->tv_nsec = tv.tv_usec*1000;
}

int local_kernel_thread(int (*fn)(void *),
		void * arg,
		unsigned long flags,
		pid_t pid);
int asm_kernel_thread(int (*fn)(void *),
		void * arg,
		unsigned long flags,
		pid_t pid);

unsigned int test_cpu_caps(void);
unsigned int test_kernel_config(void);

#define test_one_flag(src, dst, flag, message, ret) \
if (src & (1 << flag)) \
	if (!(dst & (1 << flag))) { \
		wprintk("Destination cpu does not have " message "\n"); \
		ret = 1; \
	}

static inline void
_set_normalized_timespec(struct timespec *ts, time_t sec, long nsec)
{
	while (nsec >= NSEC_PER_SEC) {
		nsec -= NSEC_PER_SEC;
		++sec;
	}
	while (nsec < 0) {
		nsec += NSEC_PER_SEC;
		--sec;
	}
	ts->tv_sec = sec;
	ts->tv_nsec = nsec;
}

static inline struct timespec
_ns_to_timespec(const nsec_t nsec)
{
	struct timespec ts;

	if (!nsec)
		return (struct timespec) {0, 0};

	ts.tv_sec = div_long_long_rem_signed(nsec, NSEC_PER_SEC, &ts.tv_nsec);
	if (unlikely(nsec < 0))
		_set_normalized_timespec(&ts, ts.tv_sec, ts.tv_nsec);

	return ts;
}
