/*
 *
 *  kernel/cpt/rst_ubc.c
 *
 *  Copyright (C) 2000-2005  SWsoft
 *  All rights reserved.
 *
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 */

#include <linux/types.h>
#include <ub/beancounter.h>

#include "cpt_obj.h"
#include "cpt_context.h"

struct user_beancounter *rst_lookup_ubc(__u64 pos, struct cpt_context *ctx)
{
	cpt_object_t *obj;

	obj = lookup_cpt_obj_bypos(CPT_OBJ_UBC, pos, ctx);
	if (obj == NULL) {
		printk(KERN_ERR "RST: unknown ub @%Lu\n", pos);
		return get_beancounter(get_exec_ub());
	}
	return get_beancounter(obj->o_obj);
}

static void restore_one_bc_parm(__u64 *dmp, struct ubparm *prm, int held)
{
	prm->barrier = (dmp[0] == CPT_NULL ? UB_MAXVALUE : dmp[0]);
	prm->limit = (dmp[1] == CPT_NULL ? UB_MAXVALUE : dmp[1]);
	if (held)
		prm->held = dmp[2];
	prm->maxheld = dmp[3];
	prm->minheld = dmp[4];
	prm->failcnt = dmp[5];
}

static int restore_one_bc(struct cpt_beancounter_image *v,
		cpt_object_t *obj, struct cpt_context *ctx)
{
	struct user_beancounter *bc;
	cpt_object_t *pobj;
	int i;

	if (v->cpt_parent != CPT_NULL) {
		pobj = lookup_cpt_obj_bypos(CPT_OBJ_UBC, v->cpt_parent, ctx);
		if (pobj == NULL)
			return -ESRCH;
		bc = get_subbeancounter_byid(pobj->o_obj, v->cpt_id, 1);
	} else {
		bc = get_exec_ub();
		while (bc->parent)
			bc = bc->parent;
		get_beancounter(bc);
	}
	if (bc == NULL)
		return -ENOMEM;
	obj->o_obj = bc;

	for (i = 0; i < UB_RESOURCES; i++)
		restore_one_bc_parm(v->cpt_parms, bc->ub_parms, 0);
	for (i = 0; i < UB_RESOURCES; i++)
		restore_one_bc_parm(v->cpt_parms + UB_RESOURCES * 6,
				bc->ub_store, 1);
	return 0;
}

int rst_undump_ubc(struct cpt_context *ctx)
{
	loff_t start, end;
	struct cpt_beancounter_image *v;
	cpt_object_t *obj;
	int err;

	err = rst_get_section(CPT_SECT_UBC, ctx, &start, &end);
	if (err)
		return err;

	while (start < end) {
		v = cpt_get_buf(ctx);
		err = rst_get_object(CPT_OBJ_UBC, start, v, ctx);
		if (err) {
			cpt_release_buf(ctx);
			return err;
		}

		obj = alloc_cpt_object(GFP_KERNEL, ctx);
		cpt_obj_setpos(obj, start, ctx);
		intern_cpt_object(CPT_OBJ_UBC, obj, ctx);

		restore_one_bc(v, obj, ctx);

		cpt_release_buf(ctx);
		start += v->cpt_next;
	}
	return 0;
}

void rst_finish_ubc(struct cpt_context *ctx)
{
	cpt_object_t *obj;

	for_each_object(obj, CPT_OBJ_UBC)
		put_beancounter(obj->o_obj);
}
