/*
 *
 *  kernel/cpt/rst_undump.c
 *
 *  Copyright (C) 2000-2005  SWsoft
 *  All rights reserved.
 *
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/file.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/pagemap.h>
#include <linux/namespace.h>
#include <linux/personality.h>
#include <linux/binfmts.h>
#include <linux/smp_lock.h>
#include <linux/ve_proto.h>
#include <linux/virtinfo.h>
#include <linux/compat.h>
#include <linux/vzcalluser.h>
#include <ub/beancounter.h>
#include <asm/desc.h>
#include <asm/unistd.h>

#include "cpt_obj.h"
#include "cpt_context.h"
#include "cpt_files.h"
#include "cpt_mm.h"
#include "cpt_process.h"
#include "cpt_socket.h"
#include "cpt_net.h"
#include "cpt_ubc.h"
#include "cpt_kernel.h"

static int rst_utsname(cpt_context_t *ctx);


struct thr_context {
	struct completion init_complete;
	struct completion task_done;
	int error;
	struct cpt_context *ctx;
	cpt_object_t	*tobj;
};

static int rst_clone_children(cpt_object_t *obj, struct cpt_context *ctx);

static int vps_rst_veinfo(struct cpt_context *ctx)
{
	int err;
	struct cpt_veinfo_image *i;
	struct ve_struct *ve;
	struct timespec delta;
	loff_t start, end;

	err = rst_get_section(CPT_SECT_VEINFO, ctx, &start, &end);
	if (err)
		goto out;

	i = cpt_get_buf(ctx);
	err = rst_get_object(CPT_OBJ_VEINFO, start, i, ctx);
	if (err)
		goto out_rel;

	ve = get_exec_env();
	ve->_shm_ctlall = i->shm_ctl_all;
	ve->_shm_ctlmax = i->shm_ctl_max;
	ve->_shm_ctlmni = i->shm_ctl_mni;

	ve->_msg_ctlmax = i->msg_ctl_max;
	ve->_msg_ctlmni = i->msg_ctl_mni;
	ve->_msg_ctlmnb = i->msg_ctl_mnb;

	BUG_ON(sizeof(ve->_sem_ctls) != sizeof(i->sem_ctl_arr));
	ve->_sem_ctls[0] = i->sem_ctl_arr[0];
	ve->_sem_ctls[1] = i->sem_ctl_arr[1];
	ve->_sem_ctls[2] = i->sem_ctl_arr[2];
	ve->_sem_ctls[3] = i->sem_ctl_arr[3];

	cpt_timespec_import(&delta, i->start_timespec_delta);
	_set_normalized_timespec(&ve->start_timespec,
			ve->start_timespec.tv_sec - delta.tv_sec,
			ve->start_timespec.tv_nsec - delta.tv_nsec);
	ve->start_jiffies -= i->start_jiffies_delta;
	// // FIXME: what???
	// // ve->start_cycles -= i->start_jiffies_delta * cycles_per_jiffy;

	err = 0;
out_rel:
	cpt_release_buf(ctx);
out:
	return err;
}

static int vps_rst_reparent_root(cpt_object_t *obj, struct cpt_context *ctx)
{
	int err;
	struct env_create_param2 param;

	ctx->cpt_jiffies64 = get_jiffies_64();
	do_gettimespec(&ctx->delta_time);

	ctx->delta_time.tv_sec -= ctx->start_time.tv_sec;
	if (ctx->start_time.tv_nsec > ctx->delta_time.tv_nsec) {
		ctx->delta_time.tv_sec--;
		ctx->delta_time.tv_nsec = 1000000000 - (ctx->start_time.tv_nsec - ctx->delta_time.tv_nsec);
	} else {
		ctx->delta_time.tv_nsec -= ctx->start_time.tv_nsec;
	}

	memset(&param, 0, sizeof(param));
	param.iptables_mask = ctx->iptables_mask;

	err = real_env_create(ctx->ve_id, VE_CREATE|VE_LOCK, 2, &param, sizeof(param));
	if (err < 0)
		eprintk_ctx("real_env_create: %d\n", err);
	get_exec_env()->jiffies_fixup = ((ctx->delta_time.tv_sec < 0) ?
		0 : timespec_to_jiffies(&ctx->delta_time)) -
		(unsigned long)(ctx->cpt_jiffies64 - ctx->virt_jiffies64);
	return err < 0 ? err : 0;
}


static int hook(void *arg)
{
	struct thr_context *thr_ctx = arg;
	struct cpt_context *ctx;
	cpt_object_t *tobj;
	struct cpt_task_image *ti;
	int err = 0;

	current->state = TASK_UNINTERRUPTIBLE;
	complete(&thr_ctx->init_complete);
	schedule();

	ctx = thr_ctx->ctx;
	tobj = thr_ctx->tobj;
	ti = tobj->o_image;

	current->fs->umask = 0;

	if (ti->cpt_pid == 1) {
		err = vps_rst_reparent_root(tobj, ctx);

		if (err) {
			rst_report_error(err, ctx);
			goto out;
		}

		memcpy(&get_exec_env()->cap_default, &ti->cpt_ecap, sizeof(kernel_cap_t));

		if (ctx->statusfile) {
			fput(ctx->statusfile);
			ctx->statusfile = NULL;
		}

		if (ctx->lockfile) {
			mm_segment_t oldfs;
			ssize_t err = -EINVAL;
			char b;

			oldfs = get_fs(); set_fs(KERNEL_DS);
			if (ctx->lockfile->f_op && ctx->lockfile->f_op->read)
				err = ctx->lockfile->f_op->read(ctx->lockfile, &b, 1, &ctx->lockfile->f_pos);
			set_fs(oldfs);
			fput(ctx->lockfile);
			ctx->lockfile = NULL;
		}

		err = vps_rst_veinfo(ctx);
		if (err) {
			eprintk_ctx("rst_veinfo: %d\n", err);
			goto out;
		}

		err = rst_utsname(ctx);
		if (err) {
			eprintk_ctx("rst_utsname: %d\n", err);
			goto out;
		}

		err = rst_root_namespace(ctx);
		if (err) {
			eprintk_ctx("rst_namespace: %d\n", err);
			goto out;
		}

		if ((err = rst_restore_net(ctx)) != 0) {
			eprintk_ctx("rst_restore_net: %d\n", err);
			goto out;
		}

		err = rst_sockets(ctx);
		if (err) {
			eprintk_ctx("rst_sockets: %d\n", err);
			goto out;
		}
		err = rst_sysv_ipc(ctx);
		if (err) {
			eprintk_ctx("rst_sysv_ipc: %d\n", err);
			goto out;
		}
	}

	do {
		if (current->user->uid != ti->cpt_user) {
			struct user_struct *u = alloc_uid(ti->cpt_user);
			if (!u) {
				eprintk_ctx("alloc_user\n");
			} else {
				switch_uid(u);
			}
		}
	} while (0);

	if ((err = rst_mm_complete(ti, ctx)) != 0) {
		eprintk_ctx("rst_mm: %d\n", err);
		goto out;
	}

	if ((err = rst_files_complete(ti, ctx)) != 0) {
		eprintk_ctx("rst_files: %d\n", err);
		goto out;
	}

	if ((err = rst_fs_complete(ti, ctx)) != 0) {
		eprintk_ctx("rst_fs: %d\n", err);
		goto out;
	}

	if ((err = rst_semundo_complete(ti, ctx)) != 0) {
		eprintk_ctx("rst_semundo: %d\n", err);
		goto out;
	}

	if ((err = rst_signal_complete(ti, ctx)) != 0) {
		eprintk_ctx("rst_signal: %d\n", err);
		goto out;
	}

	if (ti->cpt_namespace == CPT_NULL)
		exit_namespace(current);

	if (ti->cpt_personality != 0)
		__set_personality(ti->cpt_personality);

	current->set_child_tid = NULL;
	current->clear_child_tid = NULL;
	current->flags &= ~(PF_FORKNOEXEC|PF_SUPERPRIV);
	current->flags |= ti->cpt_flags&(PF_FORKNOEXEC|PF_SUPERPRIV);
	current->exit_code = ti->cpt_exit_code;
	current->pdeath_signal = ti->cpt_pdeath_signal;

	if (ti->cpt_restart.fn != CPT_RBL_0) {
		if (ti->cpt_restart.fn != CPT_RBL_NANOSLEEP
		    && ti->cpt_restart.fn != CPT_RBL_COMPAT_NANOSLEEP
		    ) {
			eprintk_ctx("unknown restart block\n");
		} else {
			current->thread_info->restart_block.fn = nanosleep_restart;
#ifdef CONFIG_X86_64
			if (!ti->cpt_64bit)
				current->thread_info->restart_block.fn = compat_nanosleep_restart;
#endif
			if (ctx->image_version != 0) {
				current->thread_info->restart_block.arg0 = ti->cpt_restart.arg0;
				current->thread_info->restart_block.arg1 = ti->cpt_restart.arg1;
				current->thread_info->restart_block.arg2 = ti->cpt_restart.arg2;
				current->thread_info->restart_block.arg3 = ti->cpt_restart.arg3;
				if (debug_level > 2) {
					ktime_t e, e1;
					struct timespec now;

					do_posix_clock_monotonic_gettime(&now);
					e = timespec_to_ktime(now);
					e1.tv64 = ((u64)current->thread_info->restart_block.arg1 << 32) | (u64) current->thread_info->restart_block.arg0;
					e = ktime_sub(e1, e);
					dprintk("rst " CPT_FID " RBL %ld/%ld %Ld\n", CPT_TID(current),
						current->thread_info->restart_block.arg1,
						current->thread_info->restart_block.arg0, e.tv64);
				}
			} else {
				struct timespec now;
				ktime_t expire;
				unsigned long val = ti->cpt_restart.arg0 -
					timespec_to_jiffies(&ctx->delta_time);
				if ((long)val <= 0)
					val = 1;
				do_posix_clock_monotonic_gettime(&now);
				expire = ktime_add_ns(timespec_to_ktime(now), (u64)val*TICK_NSEC);
				current->thread_info->restart_block.arg0 = expire.tv64 & 0xFFFFFFFF;
				current->thread_info->restart_block.arg1 = expire.tv64 >> 32;
				current->thread_info->restart_block.arg2 = ti->cpt_restart.arg1;
				current->thread_info->restart_block.arg3 = CLOCK_MONOTONIC;
			}
		}
	}

	if (thread_group_leader(current)) {
		current->signal->it_real_incr.tv64 = 0;
		if (ctx->image_version != 0) {
			ktime_add_ns(current->signal->it_real_incr, ti->cpt_it_real_incr);
		} else {
			ktime_add_ns(current->signal->it_real_incr, ti->cpt_it_real_incr*TICK_NSEC);
		}
		current->signal->it_prof_incr = ti->cpt_it_prof_incr;
		current->signal->it_virt_incr = ti->cpt_it_virt_incr;
		current->signal->it_prof_expires = ti->cpt_it_prof_value;
		current->signal->it_virt_expires = ti->cpt_it_virt_value;
	}

	err = rst_clone_children(tobj, ctx);
	if (err) {
		eprintk_ctx("rst_clone_children\n");
		goto out;
	}

	if (ti->cpt_pid == 1) {
		if ((err = rst_process_linkage(ctx)) != 0) {
			eprintk_ctx("rst_process_linkage: %d\n", err);
			goto out;
		}
		if ((err = rst_do_filejobs(ctx)) != 0) {
			eprintk_ctx("rst_do_filejobs: %d\n", err);
			goto out;
		}
		if ((err = rst_eventpoll(ctx)) != 0) {
			eprintk_ctx("rst_eventpoll: %d\n", err);
			goto out;
		}
		if ((err = rst_sockets_complete(ctx)) != 0) {
			eprintk_ctx("rst_sockets_complete: %d\n", err);
			goto out;
		}
		if ((err = rst_stray_files(ctx)) != 0) {
			eprintk_ctx("rst_stray_files: %d\n", err);
			goto out;
		}
		if ((err = rst_posix_locks(ctx)) != 0) {
			eprintk_ctx("rst_posix_locks: %d\n", err);
			goto out;
		}
		if ((err = rst_tty_jobcontrol(ctx)) != 0) {
			eprintk_ctx("rst_tty_jobcontrol: %d\n", err);
			goto out;
		}
		if ((err = rst_restore_fs(ctx)) != 0) {
			eprintk_ctx("rst_restore_fs: %d\n", err);
			goto out;
		}
	}

out:
	thr_ctx->error = err;
	lock_kernel();
	complete(&thr_ctx->task_done);

	if (!err && (ti->cpt_state & (EXIT_ZOMBIE|EXIT_DEAD))) {
		preempt_disable();
		current->exit_state = EXIT_ZOMBIE;
		write_lock_irq(&tasklist_lock);
		nr_zombie++;
		write_unlock_irq(&tasklist_lock);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,9)
		atomic_dec(&current->signal->live);
#endif
		current->flags |= PF_DEAD;
		if (!(ti->cpt_flags&PF_DEAD))
			wprintk_ctx("zombie %d,%d(%s) is not pf_dead\n", current->pid, virt_pid(current), current->comm);
		module_put(current->thread_info->exec_domain->module);
		if (current->binfmt)
			module_put(current->binfmt->module);
	} else {
		__set_current_state(TASK_UNINTERRUPTIBLE);
	}

	schedule();

	dprintk_ctx("leaked through %d/%d %p\n", current->pid, virt_pid(current), current->mm);

	module_put(THIS_MODULE);
	complete_and_exit(NULL, 0);
	return 0;
}

#if 0
static void set_task_ubs(struct cpt_task_image *ti, struct cpt_context *ctx)
{
	struct task_beancounter *tbc;

	tbc = task_bc(current);

	put_beancounter(tbc->fork_sub);
	tbc->fork_sub = rst_lookup_ubc(ti->cpt_task_ub, ctx);
	if (ti->cpt_mm_ub != CPT_NULL) {
		put_beancounter(tbc->exec_ub);
		tbc->exec_ub = rst_lookup_ubc(ti->cpt_mm_ub, ctx);
	}
}
#endif

static int create_root_task(cpt_object_t *obj, struct cpt_context *ctx,
		struct thr_context *thr_ctx)
{
	task_t *tsk;
	int pid;

	thr_ctx->ctx = ctx;
	thr_ctx->error = 0;
	init_completion(&thr_ctx->init_complete);
	init_completion(&thr_ctx->task_done);
#if 0
	set_task_ubs(obj->o_image, ctx);
#endif

	pid = local_kernel_thread(hook, thr_ctx, 0, 0);
	if (pid < 0)
		return pid;
	read_lock(&tasklist_lock);
	tsk = find_task_by_pid_ve(pid);
	if (tsk)
		get_task_struct(tsk);
	read_unlock(&tasklist_lock);
	if (tsk == NULL)
		return -ESRCH;
	cpt_obj_setobj(obj, tsk, ctx);
	thr_ctx->tobj = obj;
	return 0;
}

static int rst_basic_init_task(cpt_object_t *obj, struct cpt_context *ctx)
{
	task_t *tsk = obj->o_obj;
	struct cpt_task_image *ti = obj->o_image;

	memcpy(tsk->comm, ti->cpt_comm, sizeof(tsk->comm));
	rst_mm_basic(obj, ti, ctx);
	return 0;
}

static int make_baby(cpt_object_t *cobj,
		     struct cpt_task_image *pi,
		     struct cpt_context *ctx)
{
	unsigned long flags;
	struct cpt_task_image *ci = cobj->o_image;
	struct thr_context thr_ctx;
	task_t *tsk;
	pid_t pid;

	flags = rst_mm_flag(ci, ctx) | rst_files_flag(ci, ctx)
		| rst_signal_flag(ci, ctx) | rst_semundo_flag(ci, ctx);
	if (ci->cpt_rppid != pi->cpt_pid) {
		flags |= CLONE_THREAD|CLONE_PARENT;
		if (ci->cpt_signal != pi->cpt_signal ||
		    !(flags&CLONE_SIGHAND) ||
		    (!(flags&CLONE_VM) && pi->cpt_mm != CPT_NULL)) {
			eprintk_ctx("something is wrong with threads: %d %d %d %Ld %Ld %08lx\n",
			       (int)ci->cpt_pid, (int)ci->cpt_rppid, (int)pi->cpt_pid,
			       ci->cpt_signal, pi->cpt_signal, flags
			       );
			return -EINVAL;
		}
	}

	thr_ctx.ctx = ctx;
	thr_ctx.error = 0;
	init_completion(&thr_ctx.init_complete);
	init_completion(&thr_ctx.task_done);
	thr_ctx.tobj = cobj;

#if 0
	set_task_ubs(ci, ctx);
#endif

	pid = local_kernel_thread(hook, &thr_ctx, flags, ci->cpt_pid);
	if (pid < 0)
		return pid;

	read_lock(&tasklist_lock);
	tsk = find_task_by_pid_ve(pid);
	if (tsk)
		get_task_struct(tsk);
	read_unlock(&tasklist_lock);
	if (tsk == NULL)
		return -ESRCH;
	cpt_obj_setobj(cobj, tsk, ctx);
	thr_ctx.tobj = cobj;
	wait_for_completion(&thr_ctx.init_complete);
#ifdef CONFIG_SMP
	wait_task_inactive(cobj->o_obj);
#endif
	rst_basic_init_task(cobj, ctx);

	/* clone() increases group_stop_count if it was not zero and
	 * CLONE_THREAD was asked. Undo.
	 */
	if (current->signal->group_stop_count && (flags & CLONE_THREAD)) {
		if (tsk->signal != current->signal) BUG();
		current->signal->group_stop_count--;
	}

	wake_up_process(tsk);
	wait_for_completion(&thr_ctx.task_done);
	wait_task_inactive(tsk);

	return thr_ctx.error;
}

static int rst_clone_children(cpt_object_t *obj, struct cpt_context *ctx)
{
	int err = 0;
	struct cpt_task_image *ti = obj->o_image;
	cpt_object_t *cobj;

	for_each_object(cobj, CPT_OBJ_TASK) {
		struct cpt_task_image *ci = cobj->o_image;
		if (cobj == obj)
			continue;
		if ((ci->cpt_rppid == ti->cpt_pid && ci->cpt_tgid == ci->cpt_pid) ||
		    (ci->cpt_leader == ti->cpt_pid &&
		     ci->cpt_tgid != ci->cpt_pid && ci->cpt_pid != 1)) {
			err = make_baby(cobj, ti, ctx);
			if (err) {
				eprintk_ctx("make_baby: %d\n", err);
				return err;
			}
		}
	}
	return 0;
}

static int read_task_images(struct cpt_context *ctx)
{
	int err;
	loff_t start, end;

	err = rst_get_section(CPT_SECT_TASKS, ctx, &start, &end);
	if (err)
		return err;

	while (start < end) {
		cpt_object_t *obj;
		struct cpt_task_image *ti = cpt_get_buf(ctx);

		err = rst_get_object(CPT_OBJ_TASK, start, ti, ctx);
		if (err) {
			cpt_release_buf(ctx);
			return err;
		}
		if (ti->cpt_pid != 1 && !__is_virtual_pid(ti->cpt_pid)) {
			eprintk_ctx("BUG: pid %d is not virtual\n", ti->cpt_pid);
			cpt_release_buf(ctx);
			return -EINVAL;
		}
		obj = alloc_cpt_object(GFP_KERNEL, ctx);
		cpt_obj_setpos(obj, start, ctx);
		intern_cpt_object(CPT_OBJ_TASK, obj, ctx);
		obj->o_image = kmalloc(ti->cpt_next, GFP_KERNEL);
		if (obj->o_image == NULL) {
			cpt_release_buf(ctx);
			return -ENOMEM;
		}
		memcpy(obj->o_image, ti, sizeof(*ti));
		err = ctx->pread(obj->o_image + sizeof(*ti),
				 ti->cpt_next - sizeof(*ti), ctx, start + sizeof(*ti));
		cpt_release_buf(ctx);
		if (err)
			return err;
		start += ti->cpt_next;
	}
	return 0;
}


static int vps_rst_restore_tree(struct cpt_context *ctx)
{
	int err;
	cpt_object_t *obj;
	struct thr_context thr_ctx_root;

	err = read_task_images(ctx);
	if (err)
		return err;

	err = rst_undump_ubc(ctx);
	if (err)
		return err;

	for_each_object(obj, CPT_OBJ_TASK) {
		err = create_root_task(obj, ctx, &thr_ctx_root);
		if (err)
			return err;

		wait_for_completion(&thr_ctx_root.init_complete);
#ifdef CONFIG_SMP
		wait_task_inactive(obj->o_obj);
#endif
		rst_basic_init_task(obj, ctx);

		wake_up_process(obj->o_obj);
		wait_for_completion(&thr_ctx_root.task_done);
		wait_task_inactive(obj->o_obj);
		err = thr_ctx_root.error;
		if (err)
			return err;
		break;
	}

	return err;
}


int vps_rst_undump(struct cpt_context *ctx)
{
	int err;
	unsigned long umask;

	err = rst_open_dumpfile(ctx);
	if (err)
		return err;

#ifndef CONFIG_X86_64
	if (ctx->tasks64) {
		eprintk_ctx("Cannot restore 64 bit VE on this architecture\n");
		return -EINVAL;
	}
#endif

	umask = current->fs->umask;
	current->fs->umask = 0;

#ifdef CONFIG_VZ_CHECKPOINT_LAZY
	err = rst_setup_pagein(ctx);
#endif

	if (err == 0)
		err = vps_rst_restore_tree(ctx);

	if (err == 0)
		err = rst_restore_process(ctx);

	current->fs->umask = umask;

        return err;
}

static int rst_unlock_ve(struct cpt_context *ctx)
{
	struct ve_struct *env;

	env = get_ve_by_id(ctx->ve_id);
	if (!env)
		return -ESRCH;
	down_write(&env->op_sem);
	env->is_locked = 0;
	up_write(&env->op_sem);
	put_ve(env);
	return 0;
}

int rst_resume(struct cpt_context *ctx)
{
	cpt_object_t *obj;
	int err = 0;

	for_each_object(obj, CPT_OBJ_FILE) {
		struct file *file = obj->o_obj;

		fput(file);
	}

	rst_resume_network(ctx);

	for_each_object(obj, CPT_OBJ_TASK) {
		task_t *tsk = obj->o_obj;
		struct cpt_task_image *ti = obj->o_image;

		if (!tsk)
			continue;

		if (ti->cpt_state == TASK_UNINTERRUPTIBLE) {
			dprintk_ctx("task %d/%d(%s) is started\n", virt_pid(tsk), tsk->pid, tsk->comm);

			/* Weird... If a signal is sent to stopped task,
			 * nobody makes recalc_sigpending(). We have to do
			 * this by hands after wake_up_process().
			 * if we did this before a signal could arrive before
			 * wake_up_process() and stall.
			 */
			spin_lock_irq(&tsk->sighand->siglock);
			if (!signal_pending(tsk))
				recalc_sigpending_tsk(tsk);
			spin_unlock_irq(&tsk->sighand->siglock);

			wake_up_process(tsk);
		} else {
			if (ti->cpt_state == TASK_STOPPED ||
			    ti->cpt_state == TASK_TRACED) {
				set_task_state(tsk, ti->cpt_state);
			}
		}
		put_task_struct(tsk);
	}

	rst_unlock_ve(ctx);

#ifdef CONFIG_VZ_CHECKPOINT_LAZY
	rst_complete_pagein(ctx, 0);
#endif

	rst_finish_ubc(ctx);
	cpt_object_destroy(ctx);

        return err;
}

int rst_kill(struct cpt_context *ctx)
{
	cpt_object_t *obj;
	int err = 0;

	for_each_object(obj, CPT_OBJ_FILE) {
		struct file *file = obj->o_obj;

		fput(file);
	}

	for_each_object(obj, CPT_OBJ_TASK) {
		task_t *tsk = obj->o_obj;

		if (tsk == NULL)
			continue;

		if (tsk->exit_state == 0) {
			send_sig(SIGKILL, tsk, 1);

			spin_lock_irq(&tsk->sighand->siglock);
			sigfillset(&tsk->blocked);
			sigdelsetmask(&tsk->blocked, sigmask(SIGKILL));
			set_tsk_thread_flag(tsk, TIF_SIGPENDING);
			clear_tsk_thread_flag(tsk, TIF_FREEZE);
			if (tsk->flags & PF_FROZEN)
				tsk->flags &= ~PF_FROZEN;
			spin_unlock_irq(&tsk->sighand->siglock);

			wake_up_process(tsk);
		}

		put_task_struct(tsk);
	}

#ifdef CONFIG_VZ_CHECKPOINT_LAZY
	rst_complete_pagein(ctx, 1);
#endif

	rst_finish_ubc(ctx);
	cpt_object_destroy(ctx);

        return err;
}

static int rst_utsname(cpt_context_t *ctx)
{
	int err;
	loff_t sec = ctx->sections[CPT_SECT_UTSNAME];
	loff_t endsec;
	struct cpt_section_hdr h;
	struct cpt_object_hdr o;
	int i;

	if (sec == CPT_NULL)
		return 0;

	err = ctx->pread(&h, sizeof(h), ctx, sec);
	if (err)
		return err;
	if (h.cpt_section != CPT_SECT_UTSNAME || h.cpt_hdrlen < sizeof(h))
		return -EINVAL;

	i = 0;
	endsec = sec + h.cpt_next;
	sec += h.cpt_hdrlen;
	while (sec < endsec) {
		int len;
		char *ptr;
		err = rst_get_object(CPT_OBJ_NAME, sec, &o, ctx);
		if (err)
			return err;
		len = o.cpt_next - o.cpt_hdrlen;
		if (len > __NEW_UTS_LEN+1)
			return -ENAMETOOLONG;
		switch (i) {
		case 0:
			ptr = ve_utsname.nodename; break;
		case 1:
			ptr = ve_utsname.domainname; break;
		default:
			return -EINVAL;
		}
		err = ctx->pread(ptr, len, ctx, sec+o.cpt_hdrlen);
		if (err)
			return err;
		i++;
		sec += o.cpt_next;
	}

	return 0;
}
